	;these all used to be subroutines,
	;but I couldn't spare the stack space....
	;or the rom space...
	;or the execution time...
	;*grumble*


	;---- wipe the preserved game vars on startup ----
	MAC mac_start_game

	sei
	cld
	ldx #preserve
	lda #0
.loop:
		sta 0,x
		inx
		bne .loop

	ENDM



	;---- CLEAN_START variant -----
	MAC mac_reset_game

	ldx #preserve
	lda #0
.loop:
		dex
		txs
		pha
		bne .loop

	lda #sound_startup
	sta sound_request_high

	ENDM



	;---- game startup routine - called after the title screen ----
	MAC mac_init

	ldy #n_initvals
.loop:
		lda initvals,y
		ldx initdsts,y
		sta 0,x
		dey
		bpl .loop

	ldy mode
	lda mode_p0starty,y
	sta p0y
	lda mode_p1starty,y
	sta p1y

	lda SWCHB
	sta swchb_bak

	ldx #linebuf3
.blankloop:
		lda #>blank
		sta 11,x
		sta 13,x
		txa
		sec
		sbc #16
		tax
		bmi .blankloop

	ENDM





	;---- high-level routine to handle vblank ----
	MAC mac_do_vblank

	;-- do game logic --
	;a still = 0

	lda holeindex				;alternate x between 0 and 1,
	and #1						;this way neither player gets an advantage
	tax

	jsr check_movement
	jsr push_up
	sta WSYNC
	jsr capture_platform
	sta WSYNC
	jsr advance_score			;has an inx in it

	;inx						;switch to the other player
	txa
	and #1
	tax

	jsr check_movement
	jsr push_up
	sta WSYNC
	jsr capture_platform
	sta WSYNC
	jsr advance_score

	mac_set_plat_data		;translates the topbuf index into linebufptr
	mac_write_platform		;topbuf is the one that's off-screen, so we can write to it
	mac_release_platform	;we should also make it available for scoring

	lda linebufptr			;point linebufptr to the next one, which is on-screen
	adc #16					;c doesn't matter
	and #%10110000
	sta linebufptr
	
	mac_check_pause

	jsr set_bonus			;returns on cycle 10
							;set_bonus must be called after linebufptr is set up

	jsr set_bg_color

	mac_setup_score			;takes 13 cycles to hit a WSYNC

	jsr do_plgfx_ptrs		;this uses memory that's in the stack, so it needs to
							;go near the end of this section

	jsr do_sound

	mac_wipe_player_buffers
	mac_set_p1_buffer		;these should come after do_plgfx_ptrs
	mac_set_p0_buffer

	brk
	.byte vblank_wait

	sty VBLANK

	ENDM






	;---- high level routine to handle overscan ----
	MAC mac_do_overscan

	lda #2
	sta VBLANK

	;-- wipe playfield and player gfx --
	jsr wipe_grpx
	stx PF0
	stx PF1
	stx PF2
	stx HMCLR
	stx sound_request_low
	stx sound_request_high

	lda pause
	and #%00010000

	beq .nopause
		stx bg_color
		brk
		.byte 25					;adjust as needed....
		jmp .endpause
.nopause:	

	sta WSYNC

	;-- check collisions --
	lda holeindex					;do each player on alternating frames
	and #1
	tax
	mac_check_cx_pbl

	mac_check_cx_pp					;starts and ends with a WSYNC
	sta CXCLR

	;-- do power-ups --
	jsr run_timer					;x is preserved from before
	sta WSYNC

	ldx #1							;animate p1 here because there's a gap between the WSYNCs
	jsr animate_player

	mac_do_ai

	;-- do player 0 controls --
	ldx #0
	jsr animate_player
	sta WSYNC
	lda SWCHA
	jsr do_hcontrols
	sta WSYNC
	lda SWCHB
	asl
	jsr do_hmotion
	sta WSYNC
	lda SWCHA
	lsr
	lsr
	lsr
	lsr
	jsr check_jump
	sta WSYNC
	lda INPT4
	jsr check_fire
	sta WSYNC
	jsr drop_player

	;-- do player 1 controls --
	inx

	;-- load player 1 input --
	ldy mode
	lda mode_p1_ai_input,y
	beq .human
		ldy ai_fire
		lda ai_control
		bne .loaded
.human:
		lda SWCHA
		ldy INPT5
.loaded:
	sta p1_input
	sty p1_fire

	asl
	asl
	asl
	asl
	sta WSYNC
	jsr do_hcontrols
	sta WSYNC
	lda SWCHB
	jsr do_hmotion
	sta WSYNC
	lda p1_input
	jsr check_jump
	sta WSYNC
	lda p1_fire
	jsr check_fire
	sta WSYNC
	jsr drop_player


	;-- scroll the platforms --
	mac_scroll_platforms
.endpause:
	mac_advance_speed

	jsr check_die			;see if both the players went off the screen

	;-- see if reset was pressed --
	lda SWCHB
	and #%00000001
	bne .noreset
		jmp reset
.noreset:

	brk
	.byte overscan_wait

	jmp frame_start

	ENDM




	;---- check for pause (B&W/Color switch) ----
	MAC mac_check_pause

	lax SWCHB
	eor swchb_bak
	stx swchb_bak
	tax
	and #%00000010
	asl
	asl
	adc pause
	sta pause
	txa
	and #%00001000
	adc pause
	sta pause
	

	ENDM




	;---- write part of the next platform ----
	MAC mac_write_platform

	jsr random				;6
	ldx stage
	cmp stage_hole_probabilites,x
	ldx holeindex
	ldy holeoffsets,x
	lda holedata,x
	bcc .puthole			;2/3
		ora (linebufptr),y	;5
		sta (linebufptr),y	;6
		bne .check			;3
.puthole:
		eor #%11111111		;2
		and (linebufptr),y	;5
		sta (linebufptr),y	;6
.check:

	;-- make sure the next platform has at least one hole --
	;c set if a hole is needed
	ldy #9
	lda #%11111111
.loop:
		and (linebufptr),y
		dey
		cpy #4
		bpl .loop
	;if there are no holes,
	;a will still be %11111111
	cmp #%11111111

	ldy holeoffsets,x
	lda holedata,x
	eor #%11111111
	bcc .noforce
		;force a hole...
		and (linebufptr),y	;5
		sta (linebufptr),y	;6
.advance:
	inc holeindex
	lda holeindex
	cmp #21
	lda #1					;2
	bcc .inrange			;2/3
		sta holeindex		;3
		bcs .finish			;2

.noforce:
	pha						;7
	pla
	bcc .advance			;3

.inrange:
	nop						;2
	nop						;2
.finish:

	ENDM





	;---- figure which way the AI player will go ----
	MAC mac_do_ai

	lda p1x
	clc
	adc #playerwidth+(playerwidth/2)-pxoffset
	lsr
	lsr
	lsr
	tax
	sec
	sbc holeindex
	sta scratch
	txa
	sec
	sbc ai_memory
	sta moscratch
	sta WSYNC

	lda p1moveflags
	cmp #64
	bcs .ok
	tay				;test a=0
	bne .nomove
		ldy mode
		lda mode_p1_ai_input,y
		bmi .nomove
.ok:
	lda topline
	clc
	adc #1
	ldy topbuf
.find_plat:
		iny
		cmp p1y
		bcc .found
			sbc #platinterval
			bpl .find_plat
		bmi .nomove
.found:
	lda #0
.nomove:
	sta WSYNC
	bne .set_control

	ldx holeindex				;3
	lda buffer_addresses,y		;4

	ora holeoffsets,x			;4
	ldy holedata,x				;4
	tax							;2
	tya							;2
	and 0,x						;3
	bne .set_control			;2/3

	ldx holeindex				;3
	lda scratch					;3
	bpl .plus					;2/3

	cmp moscratch				;3
	bcc .set_control			;2/3
	eor #%11111111				;2
	adc #0						;2
	cmp moscratch				;3
	bcs .set_control			;2/3
	bcc .set_memory				;3
	
.plus:
	;37
	cmp moscratch				;3
	bcs .set_control			;2/3
	eor #%11111111				;2
	adc #1						;2
	cmp moscratch				;3
	bcc .set_control			;2/3

.set_memory:
	stx ai_memory

.set_control:
	sta WSYNC
	ldx #%11111111
	lda ai_memory
	cmp #127
	beq .finish_motion
	lda moscratch
	bne .normal_motion
		;try to line up with the hole
		lda p1x
		sec
		sbc #pxoffset
		and #%00000111
		eor #%11111111
		clc
		adc #4

.normal_motion:
	bpl .goleft
		;goright
		ldx #%11110111
		cmp #-11
		bcs .finish_motion
			lda SWCHB
			bmi .finish_motion	
				ldx #%11111011
				bne .finish_motion
.goleft:
	ldx #%11111011
	cmp #12
	bcc .finish_motion
		lda SWCHB
		bmi .finish_motion
			ldx #%11110111

.finish_motion:
	stx ai_control
	sta WSYNC

	lda #%11111111
	ldy p1y
	cpy #ai_fire_line
	adc #0				;goes to #%00000000 if y is over the line...
	sta ai_fire

	ENDM







	;---- scroll platforms ----
	;outputs: y=1 if a platform went off-screen
	MAC mac_scroll_platforms

	ldy #0
	sty scratch
	lda scrollspeed
	asl
	rol scratch
	asl
	rol scratch

	adc scrollcount
	sta scrollcount
	lda scratch
	adc topline
	sta topline

	cmp #mainlines+1
	bcc .onscreen			;2/3
		;if it scrolled off the screen, move topline down to the next one
		iny					;2			
		sbc #platinterval	;2		this sets c, so we advance topbuf below also
.continue:
		sta topline			;3
		lda topbuf			;3		
		adc #0				;2
		and #3				;2
		sta topbuf			;3
		bpl .finish

.onscreen:
	bcc .continue			;3

.finish:
	ENDM





	;---- increase the scroll speed if appropriate ----
	;input:	y=1 if speed should increase, 0 otherwise
	;		topbuf set appropriately
	MAC mac_advance_speed

	tya
	and topbuf
	lsr					;y=1, topbuf=1 or 3
	;a=0

	adc scrollspeed		;3
	sta scrollspeed		;3

	;assuming c clear
	sbc #defscrollspeed-1
	lsr
	lsr
	and #%00001111
	sta stage

	ENDM





	;---- see if a player collected a bonus ----
	;inputs:	x=player
	MAC mac_check_cx_pbl

	ldy #sound_collect
	lda CXP0FB,x
	and #%01000000			;bit 6 is set if there was a collision
	beq .nosound
		sty sound_request_high
.nosound:
	tay
	ora bonus				;this marks the bonus as collected
	sta bonus
	tya
	lsr
	lsr
	adc p0hscore,x			;using the upper 4 bits of this
	cmp #maxpower
	bcc .finish
		and #%00011111
.finish:	
	sta p0hscore,x
	
	ENDM





	;---- see if the players collided ----
	MAC mac_check_cx_pp

	sta WSYNC
	ldy mode			;3
	lda mode_pl_cx,y	;4
	and CXPPMM			;3
	bpl .nocx			;2/3

	lda p0frame			;3
	bpl .firstcx		;2/3

	lda p1x				;3
	cmp p0x				;3
	bcs .p1x_greater	;2/3
		lda p1dx		;3
		sbc #7			;2
		sta p1dx		;3

		lda p0dx		;3
		clc				;2
		adc #8			;2
		jmp .store

.firstcx:
	;18 prior
	ora #%10000000		;2
	sta p0frame			;3

	lda p0dy
	ldy p1dy
	sta p1dy
	sty p0dy

	ldy p0dx
	lda p1dx
	sty p1dx

	ldy #sound_bounce		;2
	sty sound_request_low	;3
	bne .store

.nocx:
	lda p0frame
	and #%01111111
	sta p0frame
	bpl .finish

.p1x_greater:
	lda p1dx
	adc #7
	sta p1dx

	lda p0dx
	sec
	sbc #8
.store:
	sta p0dx
.finish:
	sta WSYNC

	ENDM





	;---- prepare the mid-platform player display ----
	MAC mac_wipe_player_buffers

	ldx #linebuf3
.loop:
		lda #<blank
		sec
		sbc 3,x
		sta 10,x
		sta 12,x
		txa
		sbc #16
		tax
		bmi .loop

	ENDM




	;---- set up the mid-platform display for p0 ----
	MAC mac_set_p0_buffer

	lda p0moveflags					;3
	and #%00000111					;2
	sta WSYNC
	beq .finish						;2/3

	tay								;2
	ldx buffer_addresses,y			;4
	jsr find_plat_y					;28
	sec								;2		result might be negative...
	sbc p0y							;3
	clc								;2
	adc #playerheight				;2
	clc								;2
	adc p0gfxptr					;3
	sec								;2
	sbc 3,x							;4
	sta 10,x						;4
	lda p0gfxptr+1					;3
	sta 11,x						;4
.finish:
	sta WSYNC

	ENDM



	;---- set up the mid-platform display for p1 ----
	MAC mac_set_p1_buffer

	lda p1moveflags					;3
	and #%00000111					;2
	sta WSYNC
	beq .finish						;2/3

	tay								;2
	ldx buffer_addresses,y			;4
	jsr find_plat_y					;28
	sec								;2
	sbc p1y							;3
	clc								;2
	adc #playerheight
	clc
	adc p1gfxptr
	sec
	sbc 3,x
	sta 12,x
	lda p1gfxptr+1
	sta 13,x
.finish:
	sta WSYNC
	
	ENDM




	;---- set platform data ----
	MAC mac_set_plat_data

	ldy topbuf				;3			;this is an index to the buffer that's off-screen
	lda buffer_addresses,y	;4
	sta linebufptr			;3

	lda topline

	iny
	ldx buffer_addresses,y
	sta 0,x

	iny
	iny
	ldx buffer_addresses,y
	sec
	sbc #platinterval
	sta 0,x

	iny
	ldx buffer_addresses,y
	sbc #platinterval
	sta 0,x

	ENDM




	;---- make a platform available for scoring ----
	MAC mac_release_platform

	ldy topbuf
	lda capture_data,y
	eor #%00001111
	and captures
	sta captures

	ldy #3
	lda #free_pf_colu
	sta (linebufptr),y

	ENDM



	;---- prepare to display the score ----
	;(call this during vblank)
	;returns with x=0!
	MAC mac_setup_score

	ldx #1
	lda #100+sxoffset
	jsr position_player

	lda #scorecolubk
	sta COLUBK

	dex
	lda #20+sxoffset
	jsr position_player

	;moved to do_sound!
	;sta WSYNC
	;sta HMOVE

	ENDM





	;---- get some stuff ready for the kernel ----
	MAC mac_kernel_prep

	lda scratch
	sta COLUP0

	lda scratch+1
	sta COLUP1

	ldx #1
	lda p1x
	jsr position_player

	lda topline
	and #%00111100
	lsr
	lsr							;clears c
	tax
	lda bonus_colors,x
	sta colubonus

	lda bonus					;this points to the platform after the bonus
	adc #48
	and #%10110000
	tax
	lda #<ballvals-platheight
	sec
	sbc 3,x
	sta ballptr
	lda #>ballvals
	sta ballptr+1

	ldx #0
	lda p0x
	jsr position_player

	ENDM




	;---- do player reflections - 39 cycles ----
	MAC mac_do_player_ref

	ldx #2					;2
.loop:
		lda p0frame-1,x		;4
		lsr					;2	
		lsr					;2
		lsr					;2
		sta REFP0-1,x		;4
		dex					;2
		bne .loop			;2/3
	
	ENDM




	;---- display the score ----
	MAC mac_display_score

	sta WSYNC

	ldx #0
	stx VDELP0				;both vdels off
	stx REFP0
	stx REFP1

	lda #128
	sta HMBL				;turn this off

	lda #%00000011
	sta NUSIZ0
	sta NUSIZ1

	lda #>score_gfx
	sta digit0+1
	sta digit1+1
	sta digit2+1
	sta digit3+1
	sta digit4+1
	sta digit5+1

	ldy #0
	jsr find_digit_ptrs
	inx
	iny
	iny
	jsr find_digit_ptrs

	ldy #numberlines-1
	lda score_bk_colors,y
.loop:
		sta WSYNC		;update P0 before 29.1, 34.2, 40, after 0, 32, 37.1
						;update P1 before 56, 61.1, 66.2, after 0, 58.2, 63
						;(er, more or less...)

		sta COLUBK			;3

		lda score_colors,y	;4
		sta COLUP1			;3
		adc #p0colu-p1colu	;2		c doesn't matter
		sta COLUP0			;3

		lda (digit4),y	;5
		sta GRP0		;3 = 23
		lda (digit2),y	;5
		sta GRP0		;3 = 31		early, but the last couple bits are unused in every graphic...
		lda (digit0),y	;5
		sta GRP0		;3 = 39

		lda (digit5),y	;5
		sta GRP1		;3 = 47 
		lax (digit3),y	;5			hurray for lax!
		lda (digit1),y	;5
		stx GRP1		;3 = 60
		sta GRP1		;3 = 63

		lda score_bk_colors-1,y		;4
		dey							;2
		bpl .loop		;3 = 72

	;---- display the bonus field ----
	jsr wipe_grpx		;clears x, but not a

	sta COLUBK

	lda #%00000101
	sta NUSIZ0
	sta NUSIZ1

	lda #p0bonuscolu
	sta COLUP0

	mac_copy_buffer
	sta WSYNC

	lda #scorecolubk-2
	sta COLUBK

	lda p0hscore
	lsr
	lsr
	lsr
	lsr
	tay
	lda bonus_score_data,y
	sta GRP0

	lda #p1bonuscolu
	sta COLUP1

	lda p1hscore
	lsr
	lsr
	lsr
	lsr
	tay
	lda bonus_score_data,y
	sta GRP1

	sta WSYNC
	lda #scorecolubk-4
	sta COLUBK
	ldx #0
	jsr find_player_color

	sta WSYNC
	lda #scorecolubk-2
	sta COLUBK
	inx
	jsr find_player_color

	sta WSYNC

	;---- clean up a bit ----
	;this might go in kernel_prep instead...
	lda #scorecolubk
	sta COLUBK

	;ldx #1
	;x is still 1 from find_player_color

	stx VDELP0
	jsr wipe_grpx
	;dex
	;stx GRP0
	;stx GRP1
	stx NUSIZ0
	stx NUSIZ1

	ENDM





	;---- copy platform buffer ----
	MAC mac_copy_buffer

	ldx linebufptr
	lda 0,x					;4
	sta platform_data+0		;3
	lda 1,x
	sta platform_data+1
	lda 2,x
	sta platform_data+2
	lda 3,x
	sta platform_data+3
	lda 4,x
	sta platform_data+4
	lda 5,x
	sta platform_data+5
	lda 6,x
	sta platform_data+6
	lda 7,x
	sta platform_data+7
	lda 8,x
	sta platform_data+8
	lda 9,x
	sta platform_data+9
	lda 10,x
	sta platform_data+10
	lda 11,x
	sta platform_data+11
	lda 12,x
	sta platform_data+12
	lda 13,x
	sta platform_data+13
	
	ENDM