initdsts:
	.byte CTRLPF
	.byte VDELBL
	.byte scrollspeed
	.byte topline
	.byte p1frame
	.byte p0x
	.byte p1x
	.byte ai_memory
	.byte captures
	.byte linebuf0+1
	.byte linebuf1+1
	.byte linebuf2+1
	.byte linebuf3+1
	.byte p0jumpcnt
	.byte p1jumpcnt
	.byte p0lscore
	.byte p0hscore
	.byte p1lscore
	.byte p1hscore	
	.byte sound_request_high

initvals:
	.byte %00010000
	.byte 1
	.byte defscrollspeed
	.byte mainlines-22
	.byte 0
	.byte pxoffset+100
	.byte pxoffset+60-playerwidth
	.byte 127
	.byte %00001111			;the initial platforms are blank, so they shouldn't yield any score
	.byte <linebuf1		
	.byte <linebuf2
	.byte <linebuf3
	.byte <linebuf0
	.byte 127
	.byte 127
	.byte 0
	.byte 0
	.byte 0
	.byte 0
	.byte sound_respawn
n_initvals equ *-initvals




	;---- vertical sync... ----
do_vsync: subroutine
	VERTICAL_SYNC
	rts




	;---- calculate pointers to the score digits ----
	;inputs:	x=player (4 for the high score)
	;			y=digit pointer location (0 or 2)
find_digit_ptrs: subroutine
	lda p0lscore,x
	and #%00001111
	asl
	asl
	asl
	adc #<score_gfx
	sta digit0,y

	lda p0lscore,x
	and #%11110000
	lsr
	adc #<score_gfx
	sta digit2,y

	lda p0hscore,x
	and #%00001111
	asl
	asl
	asl
	adc #<score_gfx
	sta digit4,y
	rts





	;---- get a value out of the color cycle table ----
	;inputs:	x=player
	;outputs:	value stored in scratch [+1]
find_player_color: subroutine
	lda p0timer,x		;4
	and #%00000011		;2
	sta moscratch		;3
	lda p0state,x		;4
	asl					;2
	asl					;2
	adc moscratch		;3
	tay					;2
	lda color_cycles,y	;4
	adc color_base,x	;4
	sta scratch,x		;4
	rts					;6



	;---- power-up function: break tile ----
break_tile: subroutine
	lda p0x,x
	clc
	adc #playerwidth+(playerwidth/2)-pxoffset
	lsr
	lsr
	lsr
	tay

	lda p0moveflags,x
	and #%00000111
	beq .finish

	tax
	lda buffer_addresses,x
	ora holeoffsets,y
	tax
	lda holedata,y

	eor #%11111111
	and 0,x
	sta 0,x

.finish:
	rts
	


	;---- power-up function: swap players ----
swap_players: subroutine
	ldy p0y
	lda p1y

	cmp #deadline
	beq .finish
	cpy #deadline
	beq .finish

	sta p0y
	sty p1y

	ldy p0x
	lda p1x
	sta p0x
	sty p1x

	;old code - allowed you to swap with dead players
	;ldx #2
;.loop
	;	ldy p0x,x
	;	lda p1x,x
	;	sta p0x,x
	;	sty p1x,x
	;	dex
	;	dex
	;	bpl .loop
	;bmi wipe_ai

.finish:
	jmp wipe_ai




	;---- re-enable a dead player ----
reincarnate:
	lda #playerheight
	sta p0y,x

	lda #sound_respawn
	sta sound_request_high

	lda reincarnation_x,x
	sta p0x,x
	lda reincarnation_dx,x
	sta p0dx,x

	txa
	lsr				;a=0
	sta p0dy,x
	ror
	ror
	sta p0frame,x

	beq .noai
wipe_ai:
		lda #127
		sta ai_memory
.noai:

no_power_function:
	rts




	;---- decrease power-up timer ----
	;inputs:	x=player
	;outputs:	scratch=address of power-up function
run_timer: subroutine
	ldy #0
	lda holeindex			;cycles every 20 frames
	cmp #3					;1 or 2 - 0 is skipped
	bcs .finish
		dec p0timer,x
		bne .finish
			lda p0state,x
			sty p0state,x
			cmp #4			;power-up functions 4 and 5 might be changed based on the game mode
			bcc .skip		;i.e. swap becomes reincarnate, reincarnate becomes nothing
				ldy mode
				sbc mode_power_transform,y
.skip:
			tay
.finish:
	lda power_functions_lsb,y
	sta scratch
	lda power_functions_msb,y
	sta scratch+1

	;-- execute power-up function --
	sta WSYNC
	jmp (scratch)



	;---- see if a player pressed fire ----
	;inputs:	x=player
	;			a=fire button input
check_fire: subroutine
	and #%10000000
	bne .finish			;bit 7 clear if fire was pressed

	lda p0hscore,x
	lsr
	lsr
	lsr
	lsr
	beq .finish

	tay
	lda power_states,y
	sta p0state,x
	lda durations,y
	sta p0timer,x
		
	lda p0hscore,x
	and #%00001111
	sta p0hscore,x

	lda #sound_fire
	sta sound_request_high
.finish:
	rts




	;---- see if a player is jumping ----
	;inputs:	x=player
	;			a=joystick value
check_jump: subroutine
	ldy p0jumpcnt,x
	cpy #maxjump
	bcs .finish

	ldy #127
	lsr
	bcs .finish

	ldy p0jumpcnt,x
	lda scrollspeed
	lsr
	eor #%11111111
	adc jump_table,y
	sta p0dy,x

	iny
.finish:
	sty p0jumpcnt,x
	rts



	;---- calculate pointers to the player graphics ----
do_plgfx_ptrs: subroutine
	lda p0frame				;3
	and #%00111111			;2
	tay						;2
	lda plgfxtable,y		;4
	sta p0gfxptr			;3
	lda plgfxtable_msb,y	;4
	sta p0gfxptr+1			;3

	lda p1frame				;3
	and #%00111111			;2
	tay						;2
	lda plgfxtable,y		;4
	sta p1gfxptr			;3
	lda plgfxtable_msb,y	;4
	sta p1gfxptr+1			;3
	rts						;6



	;---- maybe place a bonus item on the screen ----
set_bonus: subroutine
	jsr random
	cmp #20
	sta WSYNC
	bcs .finish

	tax
	lda bonus
	and #%10110000			;other bits are used to flag it as collected
	beq .foo
	cmp linebufptr
.foo:
	beq .continue

.finish:
		sta WSYNC
		sta WSYNC
		nop
		nop
		rts

.continue:
	lda linebufptr
	sta bonus

	txa	
	asl
	asl
	asl
	adc #sxoffset+4			;HMOVE is done during the score
	ldx #4





	;---- position the player horizontally ----
	;input:	a = x coordinate
	;		x = player (4 for the ball)
position_player: subroutine
	sta WSYNC

	sec				;2
.pxloop:
		sbc #15		;2
		bcs .pxloop	;3/2

	eor #255		;2
	sbc #6			;2
	asl				;2
	asl				;2
	asl				;2
	asl				;2
	sta RESP0,x		;4
	sta WSYNC		;go to next line...
	sta HMP0,x		;4
	rts				;6





	;---- attribute a free platform to a player that has just passed it ----
	;inputs:	x=player
	;outputs:	c set if platform was captured
capture_platform: subroutine
	lda p0moveflags,x		;4
	beq .finish				;2/3
	cmp #64					;2
	bcs .finish				;2/3
	tay						;2 = 12

	txa						;2
	beq .noai				;2/3
		lda #127			;2
		sta ai_memory		;3 = 21
.noai:

	lda capture_data,y		;4
	and captures			;3
	bne .finish				;2/3
							;= 30

	lda capture_data,y		;4
	ora captures			;3
	sta captures			;3 = 40

	lda buffer_addresses,y	;4
	tay						;2 = 46

	lda capture_colors,x	;4
	sta 3,y					;5 = 55

	sec						;2
	rts						;6 = 63

.finish:
	clc
	rts



	;---- crappy (but fast!) random number generator ----
random: subroutine
	lda randseed		;ya = 16-bit value, a = low byte
	ldy randseed+1

	sta randseed+1		;randseed = value*256

	clc					;add the original value + 137 to that...
	adc #137
	tax
	tya
	adc randseed+1		;now ax = value*257 + 137

	sta randseed		;store the bytes in reverse!
	stx randseed+1

	rts					;return with a = some randomish number






	;---- increase the score if appropriate ----
	;inputs:	c set if score should increase
	;			x=player
	;outputs:	pxscore is affected
	;destroys:	a, flags
advance_score: subroutine
	;this allows the score to wrap, but...  isn't that part of the fun?	
	;(though I doubt anyone could get past 999...)

	inx
	bcs .nobit
		.byte $2c	;bit hides the next 2-byte command
.nobit:
		;this works because sound_capture_0 = 1
		stx sound_request_low

	sed
	lda p0lscore-1,x	;4
	adc #0				;2		<--- wonderfully convenient
	sta p0lscore-1,x	;4
	lda p0hscore-1,x	;4
	adc #0				;2
	sta p0hscore-1,x	;4
	cld
	rts




	;---- animate player ----
	;inputs:	x=player
	;outputs:	player anim count, frame affected
	;destroys:	a, flags
animate_player: subroutine
	lda p0ancnt,x
	clc
	adc #animspeed
	sta p0ancnt,x
	lda p0frame,x
	adc #0
	and #%11111011
	sta p0frame,x
	rts



	;---- see if the player has hit any platforms ----
	;inputs:	x=player
	;outputs:	moveflags loaded
	;destroys:	a, y, flags, scratch, moscratch
check_movement: subroutine
	stx scratch

	lda #0
	sta moscratch
	lda p0x,x
	sta moscratch+1

	;first check the y-coordinate...
	ldy #maxplats-1
.loop:
	lda topline

	clc
	adc #platheight+balltop

	sec
	sbc plat_heights,y
	bcc .negplat					;2/3

	sbc #platheight+balltop

	sec								;2
	sbc p0y,x						;4
	clc								;2
	adc #platheight+playerheight-1+balltop	;2
	bcc .noplat						;2/3
		cmp #balltop				;2
		tya							;2
		bcc .ballarea				;2/3
			adc topbuf				;3
			bcc .storemo			;3
.ballarea:
		;a few pixels above the platform
		adc topbuf					;3
		adc #%01000001				;2
.storemo:
		sta moscratch				;3
.continue:
	dey
	bpl .loop

	;now check the individual tiles...
	lda moscratch+1
	adc #1
	ldy moscratch
	jsr check_tile
	php

	lda moscratch+1
	clc
	adc #playerwidth-2
	ldy moscratch
	jsr check_tile

	lda #0
	rol
	plp
	rol
	beq .nopush				;2/3
		;the player gets pushed up if either side
		;hit a tile - set a flag so other routines know
		lda moscratch		;3
		ora #%10000000		;2
		tay					;2
.flagset:
	ldx scratch
	sty p0moveflags,x
	rts

.negplat:
	;waste 28
	lda (scratch,x)			;6
	lda (scratch,x)			;6
	nop						;2
.noplat:
	;waste 14
	nop 0					;3
	lda (scratch,x)			;6
	clc						;2
	bcc .continue			;3

.nopush:
	ldy moscratch			;3
	bcc .flagset			;3




check_tile: subroutine
	;input: a = x-coordinate (in player coordinates)
	;		y = platform offset (0 means player is in mid-air)
	;			bit 6 set means player is just above the platform
	;output: c if tile is solid
	;		nc,nz if hole
	;		nc,z if in mid-air
	;destroys: a,x,y,scratch

	clc							;2
	adc #playerwidth-pxoffset	;2
	lsr							;2
	lsr							;2
	lsr							;2
	tax							;2 = 12
check_tile_by_index:
	tya							;2
	asl							;2
	asl							;2
	bcs .wait1					;2/3
	asl							;2
	asl							;2
	beq .wait2					;2/3
	and #%00110000				;2
	ora holeoffsets,x			;4
	ldy holedata,x				;4 = 36

	tax							;2
	tya							;2
	and linebuf0,x				;4 = 44
.return:
	cmp #1						;2 = 46
	rts							;6 = 52

.wait1:
	;21 prior
	lda (scratch,x)				;6
.wait2:
	;27 prior
	lda (scratch,x)				;6
	lda (scratch,x)				;6
	nop 0
	lda #0						;2
	clc
	rts





	;---- alter player x-velocity based on joystick input ----
	;inputs:	a=joystick read value
	;			x=player
do_hcontrols: subroutine
	ldy p0state,x
	asl
	bcc .right
	asl
	bcc .left

	lda p0frame,x
	and #%11111100
	sta p0frame,x

	lda p0dx,x
	bmi .minus
		sbc #xdecel
		bcs .finish
		bcc .zerov

.minus:
		adc #xdecel-1
		bcc .finish
.zerov:
	lda #0

.finish:
	sta p0dx,x
	rts

.right:
	lda p0frame,x
	and #%10111111
	sta p0frame,x

	lda p0dx,x
	adc xaccels,y
	bmi .finish
	cmp maxrspeeds,y
	bcc .finish
	lda maxrspeeds,y
	bcs .finish

.left:
	lda p0frame,x
	ora #%01000000
	sta p0frame,x

	lda p0dx,x
	sec
	sbc xaccels,y
	bpl .finish
	cmp maxlspeeds,y
	bcs .finish
	lda maxlspeeds,y
	bcc .finish

	

	;---- move a player horizontally ----
	;inputs:	x=player
	;			a=difficulty switch value
do_hmotion: subroutine
	and #%10000000
	asl
	rol
	tay
	stx moscratch+1

	lda p0dx,x
	asl
	lda #0
	sbc #0
	eor #255
	sta scratch	

	lda p0dx,x
	asl
	rol scratch
	asl
	rol scratch

	clc
	adc p0fracx,x
	sta p0fracx,x
	lda scratch
	adc p0x,x

	bcc .right

	cmp leftbound1,y
	bcs .leftok
		lda edge_sounds,y
		sta sound_request_low
		lda p0dx,x
		eor vmask,y
		sec
		sbc vsub,y
		sta p0dx,x
		lda leftbound2,y
.leftok:
	clc
	pha
	adc #1
	bcc .continue

.right:
	cmp rightbound1,y
	bcc .rightok
		lda edge_sounds,y
		sta sound_request_low
		lda p0dx,x
		eor vmask,y
		sbc vsub,y
		sta p0dx,x
		lda rightbound2,y
.rightok:
	clc
	pha
	adc #playerwidth-2

.continue:
	ldy p0moveflags,x
	jsr check_tile
	pla
	ldx moscratch+1
	bcs .block

	ldy p0y,x
	cpy #deadline
	beq .block

	sta p0x,x
	rts

.block:
	lda #0
	sta p0dx,x
	sta p0fracx,x
	rts



	;---- make the player Fall Down ----
	;inputs:	x=player
	;outputs:	player y, dy vars affected
	;destroys:	a, y, flags, scratch
drop_player: subroutine

	IF PAL50
		lda holeindex
		lsr
	ELSE
		clc
	ENDIF

	lda p0dy,x
	adc #gravacc
	bvc .belowterm
		nop
		lda #127
.storevel:
	sta p0dy,x

	asl
	lda #0
	tay
	sbc #0
	eor #255
	sta scratch

	lda p0dy,x
	asl
	rol scratch
	asl
	rol scratch
	asl
	rol scratch
	sta scratch+1

	lda p0fracy,x
	sec
	sbc scratch+1
	sta p0fracy,x
	lda p0y,x
	sbc scratch

	cmp #playerheight+1		;2
	bcs .nobottom			;2/3
		sty p0fracy,x		;4
		sty p0dy,x			;4
		lda #playerheight+1	;2
.storey:
	sta p0y,x				;4
	rts

.belowterm:
	bvc .storevel	;3

.nobottom:
	nop				;2
	cmp #dieline	;2
	bcs .nostore	;2/3
	bcc .storey		;3

.nostore:
	lda (scratch,x)	;6
	rts



	;---- push the player up when appropriate ----
	;inputs:	x=player
	;outputs:	player vars affected
	;destroys:	a, y, flags
push_up: subroutine
	lda p0moveflags,x
	eor #%11000000		;2		bit 7 will be set if we need to push up (6 clear also)
	sta WSYNC
	bmi .finish			;2/3

	sta p0moveflags,x	;4
	jsr find_plat_y		;28
		
	clc					;2
	adc #playerheight+platheight	;2
	cmp #dieline		;2
	ldy #0				;2
	bcc .live			;2/3
		sty p0moveflags,x	;4
		lda #sound_die	;2
		sta sound_request_high	;3
		lda #deadline	;2
		bne .storey
.live:
	sty p0dy,x			;4
	sty p0fracy,x		;4
	sty p0jumpcnt,x		;4
.storey:
	sta p0y,x			;4
.finish:	
	rts



	;---- find the y-coordinate of a platform ----
	;inputs:	a=platform index (from moveflags)
	;			0 is taken to mean 'no platform', and will yield trash output
	;outputs:	a=y-coordinate
find_plat_y: subroutine
	and #%00000111			;2
	sec						;2
	sbc topbuf				;3
	;leaves c set in the useful cases
	tay						;2
	lda topline				;3
	sbc plat_heights-1,y	;4
	rts						;6 = 22





	;---- check if the players went off the top of the screen ----
check_die: subroutine
	sta WSYNC
	ldy #2		;number of living players
	ldx #1
.loop
		lda p0y,x
		cmp #deadline
		bcc .alive
			lda #state_ghost
			cmp p0state,x
			beq .ghost
				sta p0state,x
				lda #deadtime
				sta p0timer,x
				lda p0hscore,x
				and #%00001111
				sta p0hscore,x
.ghost:
		dey
.alive:
		dex
		bpl .loop

	tya
	bne .nofade
		ldx end_marker
		dex
		bmi .setmarker
			ldx #-1
.setmarker:
	stx end_marker

	ora sound_timer
	beq .gameover
.finish:
	sta WSYNC
	rts

.nofade:
	lda end_marker
	bpl .finish
	lda #127
	sta end_marker
	bne .finish
	

.gameover:
	;- check for a high score -
	ldx #1
.checkhs:
		;lda #%00001111			;cleared on death
		;and p0hscore,x
		;sta p0hscore,x

		lda hiscore_l
		cmp p0lscore,x
		lda hiscore_h
		and #%00001111
		sbc p0hscore,x
		bcc .seths
.endcheck:
		dex
		bpl .checkhs
	jmp reset

.seths:
	lda p0lscore,x
	sta hiscore_l
	txa
	lsr
	ror
	ora p0hscore,x
	sta hiscore_h
	jmp .endcheck





set_bg_color: subroutine
	;16
	ldy mode
	lda mode_alt_bg,y
	beq .normal
		lda scrollspeed
		and #%00000011
		tay
		lda alt_bg_colors,y
		bne .continue
.normal:
		ldy stage
		lda stage_bg_colors,y
.continue:
	sta scratch

	lax bg_color				;3
	bit end_marker				;3
	bmi .strobe					;2/3

	cmp scratch					;3
	beq .finish					;2/3
		and #%11110000			;2
		sec						;2
		sbc scratch				;3
		adc #15					;2
		sta WSYNC
		bcs .fadein				;2/3
			;fade out
			txa						;2
			and #%00001111			;2
			bne .noswitch			;2/3
				lda scratch			;3
				and #%11110000		;2
				sta bg_color		;3
				rts
.noswitch:
	dec bg_color		;5
	rts

.fadein:
	inc bg_color		;5
	rts

.strobe:
	and #%11110000
	sta bg_color
	dex
	txa
	and #%00001111
	ora bg_color
	sta bg_color

.finish:
	sta WSYNC
	rts




	;--- display the score (title screen) ---
title_score_display: subroutine
	ldy #numberlines-1
.score_loop:
		sta WSYNC		;update P0 before 42.2, 48.0, 53.1, after 0, 45.1, 50.1

		txa					;2
		adc score_colors,y	;4
		sta COLUP0			;3 = 9

		lda (digit4),y		;5
		sta GRP0			;3 = 17

		lda (scratch,x)		;6
		lda (scratch,x)		;6
		lda (scratch,x)		;6
		nop					;2 = 37

		lda (digit2),y		;5
		sta GRP0			;3 = 45
		lda (digit0),y		;5
		sta GRP0			;3 = 53

		dey					;2
		bpl .score_loop
	;jmp wipe_grpx




wipe_grpx: subroutine
	ldx #0
	stx ENABL
	stx GRP0
	stx GRP1
	rts



	;---- icon draw routine (title screen) ----
	;inputs:	a=lsb of pointer to icon graphic
do_icon: subroutine
	;-- set up icon --
	sta p0gfxptr

	lda #icon_colu
	sta COLUP0

	lda #>blank
	sta p0gfxptr+1

	ldx #0
	stx NUSIZ0
	lda #sxoffset+76
	jsr position_player

	sta WSYNC
	sta HMOVE

	;-- draw icon --
	ldy #iconheight-1
.icon:
		sta WSYNC
		lda (p0gfxptr),y
		sta GRP0
		dey
		bpl .icon
	rts




	;---- wsync for multiple lines ----
	;call by brk / .byte number_of_lines
	;returns with y=0
waste_lines: subroutine
	plp
	tsx
	inx
	dec 0,x
	lda (0,x)
	tay

.loop:
	sta WSYNC
	dey
	bne .loop
	rts